// Goertzel.cpp
/*
 * (c)2024 Forward Computing and Control Pty. Ltd.
 * NSW Australia, www.forward.com.au
 * This code is not warranted to be fit for any purpose. You may only use it at your own risk.
 * This generated code may be freely used for both private and commercial use
 * provided this copyright is maintained.
 */

#include "Arduino.h"
#include "Goertzel.h"

#define GZ_PI  (3.14159265358979323846)

// uncomment this for low level debug
//#define DEBUG_GzFt 

Print *Goertzel::debugPtr = NULL;
float Goertzel::scaleFactor = 1;
Print *Goertzel::gzDebugPtr = NULL;

void Goertzel::setPrintScaleFactor(float _scaleFactor) {
  if (_scaleFactor <= 0) {
    scaleFactor = 1;
  } else {
    scaleFactor = _scaleFactor;
  }
}

float Goertzel::getPrintScaleFactor() {
  return scaleFactor;
}

void Goertzel::setDebugOut(Print* out) {
  debugPtr = out;
  #ifdef DEBUG_GzFt
  gzDebugPtr = out;
  #endif
}


void Goertzel::GzFt_32bit(int32_t *data, size_t N, float sampleFreq, float *result, size_t result_size, Print* outPtr) {
  if (!data) {
    return; //nothing todo
  }
  if (result_size == 0) {
    return; // over flow
  }
  if (N == 0) {
    return; // nothing to do
  }
  float deltaK = ((float)N) / 2.0 / result_size; // this many points in range k = 0 to (N/2 -1)
  float k = 0.0;
  if (outPtr) {
    outPtr->print("K,Freq,Mag*"); outPtr->println(getPrintScaleFactor(),6);
  }
  for (size_t i = 0; i < result_size; i++) {
    float mag = GzFt_32bit_float(data, N, sampleFreq, k);
    if (result) {
      result[i] = mag;
    }
    if (outPtr) {
      outPtr->print(k);
      outPtr->print(',');
      outPtr->print(idxToFreq(k, sampleFreq, N));
      outPtr->print(',');
      outPtr->print(mag*scaleFactor, 4);
      outPtr->println();
    }
    k += deltaK;
  }
}

// input sample data and len, and sample freq, and the freq to get fourier transform for.
float Goertzel::GzFt_16bit_float(int16_t *data, size_t N, float sampleFreq, float k) {
  // size_t k = freq / (((float)sampleFreq) / N); //  sampleFreq/len is delta freq so freqIdx = freq/ (sampleFreq/len)
  if (k > N / 2) {
    k = N / 2;
  }
  if (gzDebugPtr) {
    //gzDebugPtr->print(" freq: "); gzDebugPtr->print( freq);
    gzDebugPtr->print(" DeltaF: "); gzDebugPtr->print( ((float)sampleFreq) / N, 4);
    gzDebugPtr->print(" k: "); gzDebugPtr->print(k); gzDebugPtr->println();
    gzDebugPtr->print(" pi: "); gzDebugPtr->println(GZ_PI , 6);
  }
  // calculated constants
  float A = 2 * GZ_PI * k / N;
  float B = 2 * cos(A);
  float C_real = cos(A);
  float C_imag = -sin(A);
  float D_exp = 2 * GZ_PI * k / N * (N - 1);
  float D_real = cos(D_exp);
  float D_imag = -sin(D_exp);

  if (gzDebugPtr) {
    gzDebugPtr->print(" A: "); gzDebugPtr->println(A, 6);
    gzDebugPtr->print(" B: "); gzDebugPtr->println(B, 6);
    gzDebugPtr->print(" C_real: "); gzDebugPtr->println(C_real, 6);
    gzDebugPtr->print(" C_imag: "); gzDebugPtr->println(C_imag, 6);
    gzDebugPtr->print(" D_real: "); gzDebugPtr->println(D_real, 6);
    gzDebugPtr->print(" D_imag: "); gzDebugPtr->println(D_imag, 6);
  }
  float s0 = 0;
  float s1 = 0;
  float s2 = 0;

  for (size_t i = 0; i < N - 1; i++) { // i.e 0 to (N-2) inclusive
    s0 = data[i]  + B * s1 - s2;
    s2 = s1;
    s1 = s0;
    //    if (gzDebugPtr) {
    //    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0,4);
    //    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1,4);
    //    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2,4);
    //    }
  }
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1*scaleFactor, 4);
    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2*scaleFactor, 4);
  }

  s0 = data[N - 1] + B * s1 - s2;
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
  }
  float y_real = s0 - s1 * C_real;
  float y_imag = s1 * C_imag;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_real: "); gzDebugPtr->println(y_real*scaleFactor, 6);
    gzDebugPtr->print(" y_imag: "); gzDebugPtr->println(y_imag*scaleFactor, 6);
  }
  // yD = y*D
  // yD_r = y_r * D_y - y_i * D_i;
  // yD_i = y_r * D_i + y_i * D_r;
  float y_D_real = y_real * D_real - (y_imag * D_imag);
  float y_D_imag = y_real * D_imag + y_imag * D_real;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_D_real: "); gzDebugPtr->println(y_D_real*scaleFactor, 6);
    gzDebugPtr->print(" y_D_imag: "); gzDebugPtr->println(y_D_imag*scaleFactor, 6);
  }

  float mag = sqrt((float)(y_D_real * y_D_real + y_D_imag * y_D_imag)); // sqrt of float result only

  return mag;

}
// input sample data and len, and sample freq, and the freq to get fourier transform for.
float Goertzel::GzFt_16bit_int(int16_t *data, size_t N, float sampleFreq, size_t k) {
  // size_t k = freq / (((float)sampleFreq) / N); //  sampleFreq/len is delta freq so freqIdx = freq/ (sampleFreq/len)
  if (k > N / 2) {
    k = N / 2;
  }
  if (gzDebugPtr) {
    //gzDebugPtr->print(" freq: "); gzDebugPtr->print( freq);
    gzDebugPtr->print(" DeltaF: "); gzDebugPtr->print( ((float)sampleFreq) / N, 4);
    gzDebugPtr->print(" k: "); gzDebugPtr->print(k); gzDebugPtr->println();
    gzDebugPtr->print(" pi: "); gzDebugPtr->println(GZ_PI , 6);
  }
  // calculated constants
  float A = 2 * GZ_PI * k / N;
  float B = 2 * cos(A);
  float C_real = cos(A);
  float C_imag = -sin(A);

  if (gzDebugPtr) {
    gzDebugPtr->print(" A: "); gzDebugPtr->println(A, 6);
    gzDebugPtr->print(" B: "); gzDebugPtr->println(B, 6);
    gzDebugPtr->print(" C_real: "); gzDebugPtr->println(C_real, 6);
    gzDebugPtr->print(" C_imag: "); gzDebugPtr->println(C_imag, 6);
  }
  float s0 = 0;
  float s1 = 0;
  float s2 = 0;

  for (size_t i = 0; i < N; i++) {
    s0 = data[i]  + B * s1 - s2;
    s2 = s1;
    s1 = s0;
    //    if (gzDebugPtr) {
    //    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0,4);
    //    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1,4);
    //    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2,4);
    //    }
  }
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1*scaleFactor, 4);
    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2*scaleFactor, 4);
  }
  s0 = B * s1 - s2;
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
  }
  float y_real = s0 - s1 * C_real;
  float y_imag = s1 * C_imag;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_real: "); gzDebugPtr->println(y_real*scaleFactor, 6);
    gzDebugPtr->print(" y_imag: "); gzDebugPtr->println(y_imag*scaleFactor, 6);
  }

  float mag = sqrt(y_real * y_real + y_imag * y_imag);

  return mag;
}

// input sample data and len, and sample freq, and the freq to get fourier transform for.
double Goertzel::GzFt_32bit_double(int32_t *data, size_t N, float sampleFreq, float k) {
  // size_t k = freq / (((float)sampleFreq) / N); //  sampleFreq/len is delta freq so freqIdx = freq/ (sampleFreq/len)
  if (k > N / 2) {
    k = N / 2;
  }
  if (gzDebugPtr) {
    //gzDebugPtr->print(" freq: "); gzDebugPtr->print( freq);
    gzDebugPtr->print(" DeltaF: "); gzDebugPtr->print( ((float)sampleFreq) / N, 4);
    gzDebugPtr->print(" k: "); gzDebugPtr->print(k); gzDebugPtr->println();
    gzDebugPtr->print(" pi: "); gzDebugPtr->println(GZ_PI , 6);
  }
  // calculated constants
  double A = 2 * GZ_PI * k / N;
  double B = 2 * cos(A);
  double C_real = cos(A);
  double C_imag = -sin(A);
  double D_exp = 2 * GZ_PI * k / N * (N - 1);
  double D_real = cos(D_exp);
  double D_imag = -sin(D_exp);

  if (gzDebugPtr) {
    gzDebugPtr->print(" A: "); gzDebugPtr->println(A, 6);
    gzDebugPtr->print(" B: "); gzDebugPtr->println(B, 6);
    gzDebugPtr->print(" C_real: "); gzDebugPtr->println(C_real, 6);
    gzDebugPtr->print(" C_imag: "); gzDebugPtr->println(C_imag, 6);
    gzDebugPtr->print(" D_real: "); gzDebugPtr->println(D_real, 6);
    gzDebugPtr->print(" D_imag: "); gzDebugPtr->println(D_imag, 6);
  }
  double s0 = 0;
  double s1 = 0;
  double s2 = 0;

  for (size_t i = 0; i < N - 1; i++) { // i.e 0 to (N-2) inclusive
    s0 = data[i]  + B * s1 - s2;
    s2 = s1;
    s1 = s0;
        if (gzDebugPtr) {
        gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor,4);
        gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1*scaleFactor,4);
        gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2*scaleFactor,4);
        }
  }
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1*scaleFactor, 4);
    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2*scaleFactor, 4);
  }

  s0 = data[N - 1] + B * s1 - s2;
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->println(s0*scaleFactor, 4);
  }
  double y_real = s0 - s1 * C_real;
  double y_imag = s1 * C_imag;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_real: "); gzDebugPtr->println(y_real*scaleFactor, 6);
    gzDebugPtr->print(" y_imag: "); gzDebugPtr->println(y_imag*scaleFactor, 6);
  }
  // yD = y*D
  // yD_r = y_r * D_y - y_i * D_i;
  // yD_i = y_r * D_i + y_i * D_r;
  double y_D_real = y_real * D_real - (y_imag * D_imag);
  double y_D_imag = y_real * D_imag + y_imag * D_real;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_D_real: "); gzDebugPtr->println(y_D_real*scaleFactor, 6);
    gzDebugPtr->print(" y_D_imag: "); gzDebugPtr->println(y_D_imag*scaleFactor, 6);
  }

  double mag = sqrt((double)(y_D_real * y_D_real + y_D_imag * y_D_imag)); // sqrt of double result only

  return mag;

}

// calculates FT for startFreq to endFreq at 1/4 freq steps.
// and returns max magnitude
double Goertzel::maxMag32bitQuarterStepsDouble(float startFreq, float endFreq, int32_t *data, size_t datasize, float sampleFreq, float *maxFreq) {
  size_t startK = freqToIdx_int(startFreq, sampleFreq, datasize); 
  size_t endK = freqToIdx_int(endFreq, sampleFreq, datasize); 
  if (startK > endK) {
    size_t temp = endK;
    endK = startK;
    startK = temp;
  }
  float kMax = startK;
  size_t numMags = endK - startK + 1;
  if (debugPtr) {
    debugPtr->print(" startK:"); debugPtr->print(startK);
    debugPtr->print(" endK:"); debugPtr->println(endK);
    debugPtr->print(" startF:"); debugPtr->print(idxToFreq(startK, sampleFreq, datasize));
    debugPtr->print(" endF:"); debugPtr->println(idxToFreq(endK, sampleFreq, datasize));
    debugPtr->print("numMags:"); debugPtr->println(numMags);
  }
  size_t maxIdx = (numMags-1)*4+1; // skip the quarter steps after end freq
  double mags[maxIdx] = {0};
  float idx[maxIdx] = {0};
  double maxMag = 0;
  double gzMag = 0;
  unsigned long ms = millis();
  size_t i = 0;
  for (size_t k = 0; k < numMags; k++) {
    idx[i] = k+startK;
    mags[i] = GzFt_32bit_double(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
    idx[i] = k + 0.25 + startK;
    mags[i] = GzFt_32bit_double(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
    idx[i] = k + 0.5 + startK;
    mags[i] = GzFt_32bit_double(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
    idx[i] = k + 0.75 + startK;
    mags[i] = GzFt_32bit_double(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
  }
  if (debugPtr) {
    debugPtr->print("Time(ms) for "); debugPtr->print(numMags); debugPtr->print(" GZ_ffts: "); debugPtr->print(millis() - ms); debugPtr->println("ms"); // 96ms
    debugPtr->print("Max Magnitude over "); debugPtr->print(startK * sampleFreq / datasize); debugPtr->print("Hz to ");
    debugPtr->print((startK + numMags - 1) * sampleFreq / datasize); debugPtr->print("Hz is "); debugPtr->print(maxMag*scaleFactor);
    debugPtr->print(" at "); debugPtr->println(idxToFreq(kMax, sampleFreq, datasize));
    debugPtr->println("Freq,Mag");
    for (size_t i = 0; i < maxIdx; i++) {
      debugPtr->print(idx[i]); debugPtr->print(',');
      debugPtr->print(idxToFreq(idx[i], sampleFreq, datasize)); //    debugPtr->print(k);
      debugPtr->print(","); debugPtr->println(mags[i]*scaleFactor);
    }
  }
  if (maxFreq) {
      *maxFreq = idxToFreq(kMax, sampleFreq, datasize);
  }
  return maxMag;
}


// calculates FT for startFreq to endFreq at 1/4 freq steps.
// and returns max magnitude
float Goertzel::maxMag32bitQuarterSteps(float startFreq, float endFreq, int32_t *data, size_t datasize, float sampleFreq, float *maxFreq) {
  size_t startK = freqToIdx_int(startFreq, sampleFreq, datasize); 
  size_t endK = freqToIdx_int(endFreq, sampleFreq, datasize); 
  if (startK > endK) {
    size_t temp = endK;
    endK = startK;
    startK = temp;
  }
  float kMax = startK;
  size_t numMags = endK - startK + 1;
  if (debugPtr) {
    debugPtr->print(" startK:"); debugPtr->print(startK);
    debugPtr->print(" endK:"); debugPtr->println(endK);
    debugPtr->print(" startF:"); debugPtr->print(idxToFreq(startK, sampleFreq, datasize));
    debugPtr->print(" endF:"); debugPtr->println(idxToFreq(endK, sampleFreq, datasize));
    debugPtr->print("numMags:"); debugPtr->println(numMags);
  }
  
  size_t maxIdx = (numMags-1)*4+1; // skip the quarter steps after end freq

  float mags[maxIdx] = {0};
  float idx[maxIdx] = {0};
  float maxMag = 0;
  float gzMag = 0;
  unsigned long ms = millis();
  size_t i = 0;
  for (size_t k = 0; k < numMags; k++) {
    idx[i] = k+startK;
    mags[i] = GzFt_32bit_float(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
    idx[i] = k + 0.25 + startK;
    mags[i] = GzFt_32bit_float(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
    idx[i] = k + 0.5 + startK;
    mags[i] = GzFt_32bit_float(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
    idx[i] = k + 0.75 + startK;
    mags[i] = GzFt_32bit_float(data, datasize, sampleFreq, idx[i]);
    if (mags[i] > maxMag) {
      maxMag = mags[i];
      kMax = idx[i];
    }
    i++;
    if (i>=maxIdx) {
        break;
    }
  }
  if (debugPtr) {
    debugPtr->print("Time(ms) for "); debugPtr->print(numMags); debugPtr->print(" GZ_ffts: "); debugPtr->print(millis() - ms); debugPtr->println("ms"); // 96ms
    debugPtr->print("Max Magnitude over "); debugPtr->print(startK * sampleFreq / datasize); debugPtr->print("Hz to ");
    debugPtr->print((startK + numMags - 1) * sampleFreq / datasize); debugPtr->print("Hz is "); debugPtr->print(maxMag*scaleFactor);
    debugPtr->print(" at "); debugPtr->println(idxToFreq(kMax, sampleFreq, datasize));
    debugPtr->println("Freq,Mag");
    for (size_t i = 0; i < maxIdx; i++) {
      debugPtr->print(idx[i]); debugPtr->print(',');
      debugPtr->print(idxToFreq(idx[i], sampleFreq, datasize)); //    debugPtr->print(k);
      debugPtr->print(","); debugPtr->println(mags[i]*scaleFactor);
    }
  }
  if (maxFreq) {
      *maxFreq = idxToFreq(kMax, sampleFreq, datasize);
  }
  return maxMag;
}


// input sample data and len, and sample freq, and the freq to get fourier transform for.
float Goertzel::GzFt_32bit_float(int32_t *data, size_t N, float sampleFreq, float k) {
  // size_t k = freq / (((float)sampleFreq) / N); //  sampleFreq/len is delta freq so freqIdx = freq/ (sampleFreq/len)
  if (k > N / 2) {
    k = N / 2;
  }
  if (gzDebugPtr) {
    //gzDebugPtr->print(" freq: "); gzDebugPtr->print( freq);
    gzDebugPtr->print(" DeltaF: "); gzDebugPtr->print( ((float)sampleFreq) / N, 4);
    gzDebugPtr->print(" k: "); gzDebugPtr->print(k); gzDebugPtr->println();
    gzDebugPtr->print(" pi: "); gzDebugPtr->println(GZ_PI , 6);
  }
  // calculated constants
  float A = 2 * GZ_PI * k / N;
  float B = 2 * cos(A);
  float C_real = cos(A);
  float C_imag = -sin(A);
  float D_exp = 2 * GZ_PI * k / N * (N - 1);
  float D_real = cos(D_exp);
  float D_imag = -sin(D_exp);

  if (gzDebugPtr) {
    gzDebugPtr->print(" A: "); gzDebugPtr->println(A, 6);
    gzDebugPtr->print(" B: "); gzDebugPtr->println(B, 6);
    gzDebugPtr->print(" C_real: "); gzDebugPtr->println(C_real, 6);
    gzDebugPtr->print(" C_imag: "); gzDebugPtr->println(C_imag, 6);
    gzDebugPtr->print(" D_real: "); gzDebugPtr->println(D_real, 6);
    gzDebugPtr->print(" D_imag: "); gzDebugPtr->println(D_imag, 6);
  }
  float s0 = 0;
  float s1 = 0;
  float s2 = 0;

  for (size_t i = 0; i < N - 1; i++) { // i.e 0 to (N-2) inclusive
    s0 = data[i]  + B * s1 - s2;
    s2 = s1;
    s1 = s0;
    //    if (gzDebugPtr) {
    //    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0,4);
    //    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1,4);
    //    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2,4);
    //    }
  }
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1*scaleFactor, 4);
    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2*scaleFactor, 4);
  }

  s0 = data[N - 1] + B * s1 - s2;
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
  }
  float y_real = s0 - s1 * C_real;
  float y_imag = s1 * C_imag;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_real: "); gzDebugPtr->println(y_real*scaleFactor, 6);
    gzDebugPtr->print(" y_imag: "); gzDebugPtr->println(y_imag*scaleFactor, 6);
  }
  // yD = y*D
  // yD_r = y_r * D_y - y_i * D_i;
  // yD_i = y_r * D_i + y_i * D_r;
  float y_D_real = y_real * D_real - (y_imag * D_imag);
  float y_D_imag = y_real * D_imag + y_imag * D_real;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_D_real: "); gzDebugPtr->println(y_D_real*scaleFactor, 6);
    gzDebugPtr->print(" y_D_imag: "); gzDebugPtr->println(y_D_imag*scaleFactor, 6);
  }

  float mag = sqrt((float)(y_D_real * y_D_real + y_D_imag * y_D_imag)); // sqrt of float result only

  return mag;

}
// input sample data and len, and sample freq, and the freq to get fourier transform for.
float Goertzel::GzFt_32bit_int(int32_t *data, size_t N, float sampleFreq, size_t k) {
  // size_t k = freq / (((float)sampleFreq) / N); //  sampleFreq/len is delta freq so freqIdx = freq/ (sampleFreq/len)
  if (k > N / 2) {
    k = N / 2;
  }
  if (gzDebugPtr) {
    //gzDebugPtr->print(" freq: "); gzDebugPtr->print( freq);
    gzDebugPtr->print(" DeltaF: "); gzDebugPtr->print( ((float)sampleFreq) / N, 4);
    gzDebugPtr->print(" k: "); gzDebugPtr->print(k); gzDebugPtr->println();
    gzDebugPtr->print(" pi: "); gzDebugPtr->println(GZ_PI , 6);
  }
  // calculated constants
  float A = 2 * GZ_PI * k / N;
  float B = 2 * cos(A);
  float C_real = cos(A);
  float C_imag = -sin(A);

  if (gzDebugPtr) {
    gzDebugPtr->print(" A: "); gzDebugPtr->println(A, 6);
    gzDebugPtr->print(" B: "); gzDebugPtr->println(B, 6);
    gzDebugPtr->print(" C_real: "); gzDebugPtr->println(C_real, 6);
    gzDebugPtr->print(" C_imag: "); gzDebugPtr->println(C_imag, 6);
  }
  float s0 = 0;
  float s1 = 0;
  float s2 = 0;

  for (size_t i = 0; i < N; i++) {
    s0 = data[i]  + B * s1 - s2;
    s2 = s1;
    s1 = s0;
    //    if (gzDebugPtr) {
    //    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0,4);
    //    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1,4);
    //    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2,4);
    //    }
  }
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
    gzDebugPtr->print(" s1: "); gzDebugPtr->print(s1*scaleFactor, 4);
    gzDebugPtr->print(" s2: "); gzDebugPtr->println(s2*scaleFactor, 4);
  }
  s0 = B * s1 - s2;
  if (gzDebugPtr) {
    gzDebugPtr->print(" s0: "); gzDebugPtr->print(s0*scaleFactor, 4);
  }
  float y_real = s0 - s1 * C_real;
  float y_imag = s1 * C_imag;
  if (gzDebugPtr) {
    gzDebugPtr->print(" y_real: "); gzDebugPtr->println(y_real*scaleFactor, 6);
    gzDebugPtr->print(" y_imag: "); gzDebugPtr->println(y_imag*scaleFactor, 6);
  }

  float mag = sqrt(y_real * y_real + y_imag * y_imag);

  return mag;
}

// limited to 0 to sampleLen/2
float Goertzel::freqToIdx(float freq, float sampleFreq, size_t sampleLen) {
  //freq = k * sampleFreq / dataSize;
  size_t k = (freq * sampleLen) / sampleFreq;
  if (k > sampleLen / 2) {
    k = sampleLen / 2;
  }
  return k;
}
// limited to 0 to sampleLen/2 and then rounded to nearest int
float Goertzel::freqToIdx_int(float freq, float sampleFreq, size_t sampleLen) {
  //freq = k * sampleFreq / dataSize;
  size_t k = (freq * sampleLen) / sampleFreq;
  if (k > sampleLen / 2) {
    k = sampleLen / 2;
  }
  return round(k);
}

float Goertzel::idxToFreq(float k, float sampleFreq, size_t sampleLen) {
  //freq = k * sampleFreq / dataSize;
  if (k > sampleLen / 2) {
    k = sampleLen / 2;
  }
  return (k * sampleFreq) / sampleLen;
}

// down sample from 16000 to 4000 by 4 point moving average and taking every 4th point of the result
// ie. just average 4 points and use that
// returns new size, data is overwritten with result
size_t Goertzel::downSample16bitBy4(int16_t *data, size_t len) {
  size_t newSize = len >> 2;  // skips last samples if not whole 4
  size_t i = 0;
  for (size_t k = 0; k < newSize; k++) {
    uint32_t avg = data[i++];
    for (size_t j = 0; j < 3; j++) {
      avg += data[i++]; // add another 3
    }
    data[k] = (uint16_t)(avg >> 2); // div by 4 for result
  }
  return newSize;
}

// down sample from 16000 to 4000 by 4 point moving average and taking every 4th point of the result
// ie. just average 4 points and use that
// returns new size, data is overwritten with result
size_t Goertzel::downSample16bitBy8(int16_t *data, size_t len) {
  size_t newSize = len >> 3;  // skips last samples if not whole 8
  size_t i = 0;
  for (size_t k = 0; k < newSize; k++) {
    uint32_t avg = data[i++];
    for (size_t j = 0; j < 7; j++) {
      avg += data[i++]; // add another 7
    }
    data[k] = (uint16_t)(avg >> 3); // div by 8 for result
  }
  return newSize;
}

// down sample from 16000 to 4000 by 4 point moving average and taking every 4th point of the result
// ie. just average 4 points and use that
// returns new size, data is overwritten with result
size_t Goertzel::downSample32bitBy4(int32_t *data, size_t len) {
  size_t newSize = len >> 2;  // skips last samples if not whole 4
  size_t i = 0;
  for (size_t k = 0; k < newSize; k++) {
    uint64_t avg = data[i++];
    for (size_t j = 0; j < 3; j++) {
      avg += data[i++]; // add another 3
    }
    data[k] = (uint32_t)(avg >> 2); // div by 4 for result
  }
  return newSize;
}

// down sample from 16000 to 4000 by 4 point moving average and taking every 4th point of the result
// ie. just average 4 points and use that
// returns new size, data is overwritten with result
size_t Goertzel::downSample32bitBy8(int32_t *data, size_t len) {
  size_t newSize = len >> 3; // skips last samples if not whole 8
  size_t i = 0;
  for (size_t k = 0; k < newSize; k++) {
    uint64_t avg = data[i++];
    for (size_t j = 0; j < 7; j++) {
      avg += data[i++]; // add another 7
    }
    data[k] = (uint32_t)(avg >> 3); // div by 8 for result
  }
  return newSize;
}
