#ifndef GOERTZEL_H
#define GOERTZEL_H
// Goertzel.h
/*
 * (c)2024 Forward Computing and Control Pty. Ltd.
 * NSW Australia, www.forward.com.au
 * This code is not warranted to be fit for any purpose. You may only use it at your own risk.
 * This generated code may be freely used for both private and commercial use
 * provided this copyright is maintained.
 */

#include "stddef.h"
#include "stdint.h"

class Goertzel {
  public:
    // divide output prints by this number as print(float) prints ovf for numbers > 4294967040
    static void setPrintScaleFactor(float _scaleFactor);
    static float getPrintScaleFactor();

    // calculates FT for startFreq to endFreq at 1/4 freq steps.
    // and returns max magnitude (and maxFreq if non-null pointer passed)
    static float maxMag32bitQuarterSteps(float startFreq, float endFreq, int32_t *data, size_t len, float sampleFreq, float *maxFreq=NULL);
    static double maxMag32bitQuarterStepsDouble(float startFreq, float endFreq, int32_t *data, size_t datasize, float sampleFreq, float *maxFreq=NULL);
    // result_size Fourier transforms in the range k 0 to (N/2 -1);
    static void GzFt_32bit(int32_t *data, size_t N, float sampleFreq, float *result, size_t result_size, Print* outPtr = NULL);
    
    // input sample data and len, and sample freq, and the index k to get fourier transform for.
    static float GzFt_32bit_int(int32_t *data, size_t len, float sampleFreq, size_t k); // integer idx
    static float GzFt_32bit_float(int32_t *data, size_t len, float sampleFreq, float k); // fractional idx
    static double GzFt_32bit_double(int32_t *data, size_t len, float sampleFreq, float k); // fractional idx
    static float GzFt_16bit_int(int16_t *data, size_t len, float sampleFreq, size_t k); // integer idx
    static float GzFt_16bit_float(int16_t *data, size_t len, float sampleFreq, float k); // fractional idx
    
    static float freqToIdx(float freq, float sampleFreq, size_t sampleLen);
    static float freqToIdx_int(float freq, float sampleFreq, size_t sampleLen);
    static float idxToFreq(float k, float sampleFreq, size_t sampleLen);
    static size_t downSample32bitBy4(int32_t *data, size_t len);
    static size_t downSample32bitBy8(int32_t *data, size_t len);
    static size_t downSample16bitBy4(int16_t *data, size_t len);
    static size_t downSample16bitBy8(int16_t *data, size_t len);
    static void setDebugOut(Print* out);
private:
    static Print *debugPtr;
    static float scaleFactor;
    static Print *gzDebugPtr;
};

#endif
