#ifndef PARSE_WAV_H
#define PARSE_WAV_H
// parseWAV.h
/*
 * (c)2024 Forward Computing and Control Pty. Ltd.
 * NSW Australia, www.forward.com.au
 * This code is not warranted to be fit for any purpose. You may only use it at your own risk.
 * This generated code may be freely used for both private and commercial use
 * provided this copyright is maintained.
 */

#include <Arduino.h>
#include <SD.h>
#include "waveHeader.h"

class parseWAV {
  public:
    static bool parse(File &wavefile, pcm_wav_header_t &wf); // updates wave_header by reference
    static void print_header(pcm_wav_header_t &wf, Print *outPtr);

    static uint32_t getNoOfChannels(pcm_wav_header_t &wf);
    static uint32_t getSampleRate(pcm_wav_header_t &wf);
    static uint16_t getBitsPerSample(pcm_wav_header_t &wf);
    static uint16_t getBytesPerSample(pcm_wav_header_t &wf); // bits / 8
    static size_t timeToSamples(pcm_wav_header_t &wf, float t_sec); // converts time duration to no of Samples (mono or stero)
    
    // read numOfPoints of 16bit wave return number of 16bit data point read in
    static size_t read16bitData(File &file, pcm_wav_header_t &wf, int16_t *buf, size_t numOfPoints, size_t pointOffset = 0);
    
    // read numOfPoints of 32bit wave return number of 32bit data point read in
    static size_t read32bitData(File &file, pcm_wav_header_t &wf, int32_t *buf, size_t numOfPoints, size_t pointsOffset = 0);
    
    static void print32bitDataStream(File &file, pcm_wav_header_t &wf, Print *outPtr); 
    static void print16bitDataStream(File &file, pcm_wav_header_t &wf, Print *outPtr); 
    static bool isLittleEndian();
    static void setDebugOut(Print* out);
  private:
    static Print* debugPtr;

    /* Header sizes */
    enum wave_endieness {
      LITTLE_ENDIEN = 0,
      BIG_ENDIEN    = 1,
    };

    /* Header sizes */
    enum wave_fmt_size {
      SIZE_CHUNK_ID              = 4,
      SIZE_CHUNK_SIZE            = 4,
      SIZE_FORMAT                = 4,
      SIZE_SUB_CHUNK1_ID         = 4,
      SIZE_SUB_CHUNK1_SIZE       = 4,
      SIZE_AUDIO_FORMAT          = 2,
      SIZE_NBR_CHANNELS          = 2,
      SIZE_SAMPLE_RATE           = 4,
      SIZE_BYTE_RATE             = 4,
      SIZE_BLOCK_ALIGN           = 2,
      SIZE_BITS_PER_SAMPLE       = 2,
      SIZE_SUB_CHUNK2_ID         = 4,
      SIZE_SUB_CHUNK2_SIZE       = 4,

      // LIST CHUNK SPECIFIC
      SIZE_LIST_TYPE_ID          = 4,
      SIZE_LIST_INFO_ID          = 4,
      SIZE_LIST_INFO_TEXT_SIZE   = 4,

    };

};

#endif /* end of PARSE_WAV_H */
